#include "bmp.h"
#include "effects.h"

#include <stdio.h>
#include <stdlib.h>
#include <inttypes.h>

#include <time.h>
#include <unistd.h>

struct image_t {
    struct bmp_header bmp_header;
    struct bitmapinfoheader bitmapinfoheader;
    void (*effects)(char *image_data, int width, int height, int row_size);
    FILE *f;

    struct image_t *next;
};

static void init_image(struct image_t *i, FILE *f) {
    i->f = NULL;

    if (fread(&i->bmp_header, sizeof(i->bmp_header), 1, f) != 1) {
        fprintf(stderr, "Failed to read BMP header\n");
        return;
    }

    if (i->bmp_header.type != 0x4D42) {
        fprintf(stderr, "Unexpected BMP type: %04x\n", (unsigned) i->bmp_header.type);
        return;
    }

    if (fread(&i->bitmapinfoheader, sizeof(i->bitmapinfoheader), 1, f) != 1) {
        fprintf(stderr, "Failed to read BITMAPINFOHEADER\n");
        return;
    }

    if (i->bitmapinfoheader.size < 40) {
        fprintf(stderr, "Unexpected BITMAPINFOHEADER size: %u\n", (unsigned) i->bitmapinfoheader.size);
        return;
    }

    if (i->bitmapinfoheader.planes != 1) {
        fprintf(stderr, "Unexpected BITMAPINFOHEADER color planes number: %u\n", (unsigned) i->bitmapinfoheader.planes);
        return;
    }

    if (i->bitmapinfoheader.bpp != 24) {
        fprintf(stderr, "Only 24 bpp is supported, got %u\n", (unsigned) i->bitmapinfoheader.bpp);
        return;
    }

    if (i->bitmapinfoheader.compression != 0) {
        fprintf(stderr, "Image is compressed, which is not supported\n");
        return;
    }

    fprintf(stderr, "Width x height: %"PRId32" x %"PRId32"\n",
            i->bitmapinfoheader.width,
            i->bitmapinfoheader.height);

    i->f = f;
}

static void process_image(struct image_t *i) {
    int row_size = i->bitmapinfoheader.bpp / 8 * i->bitmapinfoheader.width;
    row_size = (row_size + 3) & ~0x3; // align to 4 bytes

    char *imagedata = malloc(row_size * i->bitmapinfoheader.height);
    fseek(i->f, i->bmp_header.data_offset, SEEK_SET);
    fread(imagedata, row_size, i->bitmapinfoheader.height, i->f);

    i->effects(imagedata, i->bitmapinfoheader.width, i->bitmapinfoheader.height, row_size);
    fseek(i->f, i->bmp_header.data_offset, SEEK_SET);
    for (int y = 0; y < i->bitmapinfoheader.height; y++) {
        fwrite(&imagedata[y * row_size], row_size, 1, i->f);
    }

    free(imagedata);
}

int main(int argc, char **argv) {
    system("lsb_release -a");
    fprintf(stderr, "\n");

    struct image_t *head = NULL, **last = &head;

    srand(time(NULL) ^ getpid());

    for (int i = 1; i < argc; i++) {
        struct image_t *img = malloc(sizeof(*img));

        fprintf(stderr, "# Opening %s...\n", argv[i]);
        FILE *f = fopen(argv[i], "r+b");
        if (!f) {
            return 1;
        }

        init_image(img, f);
        img->next = NULL;
        switch (rand() % 3) {
        case 0:
            img->effects = effect_wtf1;
            break;
        case 1:
            img->effects = effect_wtf2;
            break;
        case 2:
            img->effects = effect_wtf3;
            break;
        }
        *last = img;
        last = &img->next;
    }

    while (head) {
        process_image(head);
        fclose(head->f);
        struct image_t *tmp = head;
        head = head->next;
        free(tmp);
    }
}
